#include "widgets.h"

#include "util.h"

#include <SDL2/SDL_image.h>
#include <fmt/format.h>

#include <stdexcept>

PositionInfo::PositionInfo() : x{0}, y{0}, align{Alignment::LEFT} {}

Widget::Widget(SDL_Renderer* renderer) : renderer{renderer}, rect{0, 0, 0, 0} {}

Widget::~Widget() {}

void Widget::set_width(int width, bool preserve_aspect_ratio) {
  if (preserve_aspect_ratio) {
    float scale = ((float)width) / rect.w;
    rect.h = round(rect.h * scale);
  }
  rect.w = width;
}

void Widget::set_height(int height, bool preserve_aspect_ratio) {
  if (preserve_aspect_ratio) {
    float scale = ((float)height) / rect.h;
    rect.w = round(rect.w * scale);
  }
  rect.h = height;
}

void Widget::set_position(int x, int y) {
  pos.x = x;
  pos.y = y;
  recalculate_rect();
}

void Widget::set_alignment(Alignment align) {
  pos.align = align;
  recalculate_rect();
}

void Widget::recalculate_rect() {
  int x = pos.x;
  int y = pos.y;
  switch (pos.align) {
  case Alignment::LEFT:
    break;
  case Alignment::RIGHT:
    x -= rect.w;
    break;
  case Alignment::CENTER:
    x -= rect.w / 2;
    break;
  default:
    throw std::runtime_error(
        fmt::format("Unknown alignment: {}", (int)pos.align));
  }
  rect.x = x;
  rect.y = y;
}

TextureWidget::TextureWidget(SDL_Renderer* renderer,
                             std::optional<SDL_Texture*> texture)
    : Widget{renderer}, texture{texture} {
  if (texture) {
    update_dimensions_from_texture(*texture);
  }
}

TextureWidget::~TextureWidget() {
  if (texture) {
    SDL_DestroyTexture(*texture);
    texture = nullptr;
  }
}

void TextureWidget::draw() {
  if (texture) {
    SDL_RenderCopy(renderer, *texture, NULL, &rect);
  }
}

void TextureWidget::update_texture(SDL_Texture* new_texture) {
  if (texture) {
    SDL_DestroyTexture(*texture);
  }
  update_dimensions_from_texture(new_texture);
  texture = new_texture;
}

void TextureWidget::update_dimensions_from_texture(SDL_Texture* texture) {
  if (SDL_QueryTexture(texture, nullptr, nullptr, &rect.w, &rect.h) != 0) {
    throw std::runtime_error(
        fmt::format("Couldn't query texture: {}", SDL_GetError()));
  }
}

ImageWidget::ImageWidget(SDL_Renderer* renderer, const std::string& path)
    : TextureWidget{renderer, util::load_img(renderer, path)} {}

TextWidget::TextWidget(SDL_Renderer* renderer, TTF_Font* font,
                       const std::string& initial_text)
    : TextureWidget{renderer, std::nullopt}, font{font}, text{initial_text} {
  if (text != "") {
    update_texture(generate_text(text));
  }
}

TextWidget::~TextWidget() {}

void TextWidget::update_text(const std::string& new_text) {
  if (text != new_text) {
    update_texture(generate_text(new_text));
    text = new_text;
  }
}

SDL_Texture* TextWidget::generate_text(const std::string& text) {
  SDL_Surface* surf =
      TTF_RenderText_Solid(font, text.c_str(), {0xFF, 0xFF, 0XFF, 0xFF});
  if (surf == nullptr) {
    throw std::runtime_error(
        fmt::format("Unable to render text surface: {}", TTF_GetError()));
  }
  SDL_Texture* texture = SDL_CreateTextureFromSurface(renderer, surf);
  if (texture == nullptr) {
    throw std::runtime_error(fmt::format(
        "Unable to create texture from text surface: {}", SDL_GetError()));
  }
  return texture;
}